package SlackBuild::Registry::Backend::FS;
use strict;
use warnings;
use Carp;
use File::Basename;
use List::Regexp;
use File::stat;
use File::Spec;
use Fcntl ':mode';
use SlackBuild::Registry::Record;

=head1 NAME
    
SlackBuild::Registry::Backend::FS - filesystem backend for slackbuild registry

=head1 SYNOPSIS

    $reg = new SlackBuild::Registry(dir => DIRECTORY);
    $pat = new SlackBuild::Registry::Pattern(package => 'openssl',
                                             version => '1.0.2m');
    @a = $x->lookup($pat);

=head1 METHODS

=head2 new

    $x = new SlackBuild::Registry::Backend::FS(dir => DIRECTORY)
    
=cut
    
sub new {
    my $class = shift;
    my $self = bless {}, $class;
    local %_ = @_;
    if (my $dir = delete $_{dir}) {
	$self->{dir} = $dir;
    } else {
	croak "required parameter dir not present";
    }
    croak "too many arguments" if keys %_;
    $self->scan;
    return $self;
}

my @architectures = qw(i386 x86_64 arm noarch);
my @suffixes = qw(.tgz .txz);

sub scan {
    my $self = shift;
    my $pat = "*-*-*-*";
    
    my $rx = '^(?<pkg>.+)-'
	   . '(?<vpfx>.*?)'
	   . '(?<version>\d+(?:\.\d+)*.*?)'
	   . '-(?<arch>' . regexp_opt(@architectures) . ')'
	   . '-(?<build>\d+)(?<rest>[[:punct:]].*)?$';
    $self->{ls} = [sort {
	my $d;
	if ($d = ($a->package || '') cmp ($b->package || '')) {
	    $d
	} elsif ($d = $b->version <=> $a->version) {
	    $d
	} elsif ($a->arch && $b->arch
		 && ($d = $a->arch cmp $b->arch)) {
	    $d
        } else {
	    ($b->build || 1) <=> ($a->build || 1)
        }
    }
    map {
	my ($name,$path,$suffix) = fileparse($_, @suffixes);
	if ($name =~ m{$rx}) {
	    my $st = stat($_);
	    if (S_ISREG($st->mode)) {
		new SlackBuild::Registry::Record($+{pkg},
						 version => $+{version},
						 arch => $+{arch},
						 build => $+{build},
						 date => $st->mtime,
						 filename => File::Spec->abs2rel($_, $self->{dir}))
	    } else {
		()
	    }
	} else {
	    ()
	}
    } (glob File::Spec->catfile($self->{dir}, $pat))];

    $self->{index}{$self->{ls}[0]->package}[0] = 0;
    my $i;
    for ($i = 1; $i < @{$self->{ls}}; $i++) {
	unless ($self->{ls}[$i]->package eq $self->{ls}[$i-1]->package) {
	    $self->{index}{$self->{ls}[$i-1]->package}[1] = $i-1;
	    $self->{index}{$self->{ls}[$i]->package}[0] = $i;
	}
    }
    $self->{index}{$self->{ls}[$i-1]->package}[1] = $i;
}

=head2 lookup

    @a = $backend->lookup($pattern)

Returns a sorted array of SlackBuild::Registry::Record objects matching the
B<SlackBuild::Registry::Pattern> object B<$pattern>.
    
=cut
    
sub lookup {
    my ($self, $pred) = @_;
    my $pkg = $pred->package;

    my @result;
    if ($pkg) {
	if (my $idx = $self->{index}{$pkg}) {
	    @result = grep { $pred->matches($_) }
	    		@{$self->{ls}}[$idx->[0] .. $idx->[1]];
	}
    } else {
	@result = grep { $pred->matches($_) } @{$self->{ls}};
    }

    if (wantarray) {
	(@result)
    } else {
	shift @result;
    }
}

1;
